// BinPack.h - Bin packing for 2d textures

// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.02 - 04/27/13 - A couple sort implementations included, since BinPack2 expects pre-sorted data
// - v1.01 - 04/13/13 - Silenced compiler warning
// - v1.00 - 04/11/13 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - Fast, stack based recursive method for bin packing textures. TONS of room for improvement, but I
//   haven't actually had a need for it to be faster yet (it is far from a bottleneck in any way so far).
//   As for space efficiency, it's already pretty good and the improvement would be quite minimal.
// - Splits into multiple output images as needed
// - Fails if an image is bigger than the maximum output size

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

#include <vector>
#include "Rect.h"

namespace glory {

struct BinPackTex {
	BinPackTex() : rect(0, 0, 0, 0), texI(0), imageI(0) { }
	BinPackTex(int l, int t, int r, int b, int texI) : rect(l, t, r, b), texI(texI), imageI(0) { }
	Rect rect;
	int texI;
	int imageI;  // Not needed or used on input, only on output if doesn't fit in 1 image
};

inline void BinPackSort(std::vector<BinPackTex> *pData) {
	struct Sorter {
		static bool Sort(glory::BinPackTex &lhs, glory::BinPackTex &rhs) {
			// Sort large to small, not small to large
			// AREA
			return lhs.rect.GetHeight() * lhs.rect.GetWidth() > rhs.rect.GetHeight() * rhs.rect.GetWidth();
			// MAX DIMENSION
			//if (lhs.rect.GetWidth() > rhs.rect.GetWidth() && lhs.rect.GetWidth() > rhs.rect.GetHeight())
			//	return true;
			//if (lhs.rect.GetHeight() > rhs.rect.GetWidth() && lhs.rect.GetHeight() > rhs.rect.GetHeight())
			//	return true;
			//return false;
		}
	};
	std::sort(pData->begin(), pData->end(), Sorter::Sort);
};

// Returns the number of output images
// Sorting is up to the user before hand
inline int BinPack2(const std::vector<BinPackTex> &input, std::vector<BinPackTex> *pOutput, int outW, int outH, int pad) {
	struct PackNode : public BinPackTex {
		PackNode() { childI[0] = 0; childI[1] = 0; }
		PackNode(int l, int t, int r, int b, int texI) : BinPackTex(l, t, r, b, texI) { childI[0] = 0; childI[1] = 0; }
		int childI[2];
	};
	std::vector<std::vector<PackNode> > nodes(1);
	nodes[0].reserve(input.size() * 2);
	nodes[0].push_back(PackNode(0, 0, outW - 1, outH - 1, -1));

	for (size_t inI = 0; inI < input.size(); inI++) {
		const BinPackTex *pIn = &input[inI];
		if (pIn->rect.GetWidth() > outW || pIn->rect.GetHeight() > outH)
			return 0;  // fail - can not pack all textures

		for (size_t outImageI = 0; outImageI < nodes.size(); outImageI++) {
			// Insert code
			std::vector<int> stack;
			stack.push_back(-1);  // This marks that there was NO WHERE to put it and we need to start a new image!
			stack.push_back(0);
			PackNode *pNode = 0;
			while (stack.size()) {
				if (stack.back() == -1) {
					pNode = 0;
					break;
				}
				pNode = &nodes[outImageI][stack.back()];
				stack.pop_back();
				if (pNode->childI[0]) {  // implies childI[1] is also set
					// Not a leaf here

					stack.push_back(pNode->childI[1]);
					stack.push_back(pNode->childI[0]);
				}
				else {
					// A leaf - could have an image, get an image, or split
		
					if (pNode->texI != -1)  // Leaf already has an image. Done here.
						continue;
					int dw = pNode->rect.GetWidth() - (pIn->rect.GetWidth() + pad);
					int dh = pNode->rect.GetHeight() - (pIn->rect.GetHeight() + pad);
					if (dw < 0 || dh < 0)  // Texture too big.
						continue;
					if (dw == 0 && dh == 0)  // Texture just right.
						break;
					// Change pNode before calling push_back since that can actually change the pointer making pNode invalid
					pNode->childI[0] = static_cast<int>(nodes[outImageI].size());
					pNode->childI[1] = static_cast<int>(nodes[outImageI].size() + 1);
					stack.push_back(pNode->childI[0]);
					if (dw > dh) {
						PackNode n1(pNode->rect.l, pNode->rect.t, pNode->rect.l + pIn->rect.GetWidth() - 1 + pad, pNode->rect.b, -1);
						PackNode n2(pNode->rect.l + pIn->rect.GetWidth() + pad, pNode->rect.t, pNode->rect.r, pNode->rect.b, -1);
						nodes[outImageI].push_back(n1);
						nodes[outImageI].push_back(n2);
					}
					else {
						PackNode n1(pNode->rect.l, pNode->rect.t, pNode->rect.r, pNode->rect.t + pIn->rect.GetHeight() - 1 + pad, -1);
						PackNode n2(pNode->rect.l, pNode->rect.t + pIn->rect.GetHeight() + pad, pNode->rect.r, pNode->rect.b, -1);
						nodes[outImageI].push_back(n1);
						nodes[outImageI].push_back(n2);
					}
				}
			}

			if (pNode) {
				// Found an place for the texture. Otherwise, we need to start a new image to store more textures.
				pNode->texI = pIn->texI;
				pOutput->push_back(BinPackTex(pNode->rect.l, pNode->rect.t, pNode->rect.r - pad, pNode->rect.b - pad, pNode->texI));
				pOutput->back().imageI = static_cast<int>(outImageI);
				break;
			}
			else {
				if (outImageI == nodes.size() - 1) {
					nodes.resize(nodes.size() + 1);
					nodes[outImageI + 1].reserve(input.size() * 2);
					nodes[outImageI + 1].push_back(PackNode(0, 0, outW - 1, outH - 1, -1));
				}
			}
		}
	}

	return static_cast<int>(nodes.size());
}

}  // namespace glory
