// Cam2dZoom.h - Basic 2d camera supporting a single layer of panning and 
// zooming in 2d environments.
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.01 - 07/18/12 - BUGFIX: Changing limit with SetLimits now applies those limits to current state
//                    - BUGFIX: Zoom limits are always checked when limits are enforced
// - v1.00 - 05/31/12 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - Correctly supports panning and zooming with world limits automatically. For instance,
//   If you zoom *out* while panned far right, the camera will correctly adjust 
//   the pan position in order to preserve world limits while zooming out.
// - PixToWorld and WorldToPix do not just convert n pixels to m world-units. Instead, they
//   convert absolute positions of one coordinate system to the other. The pixel at 0, 0 for
//   instance will coorespond to a different world-unit coordinate depending on *where* the
//   camera is panned at the time.

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

#include <math.h>

namespace glory {

// Intended for a true 2d plane to be zoomed and scrolled on. NOT intended for those kinds of multi-level
// 2d planes that can be brought forth to the foreground.
class Cam2dZoom {
public:
	Cam2dZoom() : zoom(1.0f), cx(0.0f), cy(0.0f), viewW(640.0f), viewH(480.0f), pixW(640), pixH(480),
		limL(-FLT_MAX), limT(-FLT_MAX), limR(FLT_MAX), limB(FLT_MAX), limZ(.00001f) { }
	~Cam2dZoom() { }

	// Pixel size of the viewport
	void SetViewportPix(int w, int h) { pixW = w; pixH = h; }

	// How many units of the world should the viewport be at zoom level 1.0.
	void SetViewport(float w, float h) { viewW = w; viewH = h; }

	// Note this does NOT convert units - rather it converts from one coordinate to the other.
	// The output will depend on the *position* of the camera - not just pix-to-world scale.
	void PixToWorld(float px, float py, float *wx, float *wy) const {
		if (wx) {
			float scaleX = (viewW / pixW) / zoom;
			*wx = (px - pixW * .5f) * scaleX + cx;
		}
		if (wy) {
			float scaleY = (viewH / pixH) / zoom;
			*wy = (py - pixH * .5f) * scaleY + cy;
		}
	}

	// Note this does NOT convert units - rather it converts from one coordinate to the other.
	// The output will depend on the *position* of the camera - not just world-to-pix scale.
	void WorldToPix(float wx, float wy, float *px, float *py) const {
		if (px) {
			float scaleX = zoom * pixW / viewW;
			*px = ((wx - cx) * scaleX) + pixW * .5f;
		}
		if (py) {
			float scaleY = zoom * pixH / viewH;
			*py = ((wy - cy) * scaleY) + pixH * .5f;
		}
	}

	void GetWorldToPixScale(float *px, float *py) const {
		if (px)
			*px = (pixW / viewW) * zoom;
		if (py)
			*py = (pixH / viewH) * zoom;
	}

	void GetPixToWorldScale(float *wx, float *wy) const {
		if (wx)
			*wx = viewW / (zoom * pixW);
		if (wy)
			*wy = viewH / (zoom * pixH);
	}

	// In world units
	void SetLimits(float left, float top, float right, float bottom) {
		limL = left;
		limT = top;
		limR = right;
		limB = bottom;
		float xr = (limR - limL) / viewW;
		float yr = (limB - limT) / viewH;
		if (xr > yr)  // Fit width, spacing on height?
			limZ = 1.0f / xr;
		else
			limZ = 1.0f / yr;
		Limit_();
	}

	// The x and y movements are given in pixel units.
	void MovePix(float x, float y) {
		float scaleX = viewW / (zoom * pixW);
		float scaleY = viewH / (zoom * pixH);
		cx += x * scaleX;
		cy += y * scaleY;
		Limit_();
	}

	// In world units
	void Move(float x, float y) { cx += x; cy += y; Limit_(); }

	// In world units
	void SetPos(float x, float y) { cx = x; cy = y; Limit_(); }

	void SetZoom(float scale) { zoom = scale; Limit_(); }
	
	float GetZoom() const { return zoom; }

	// In world units
	float GetX() const { return cx; }

	// In world units
	float GetY() const { return cy; }

	// In world units
	float GetLeft() const { return cx - (viewW * .5f) / zoom; }

	// In world units
	float GetRight() const { return cx + (viewW * .5f) / zoom; }

	// In world units
	float GetTop() const { return cy - (viewH * .5f) / zoom; }

	// In world units
	float GetBottom() const { return cy + (viewH * .5f) / zoom; }

	// In world units
	float GetWidth() const { return viewW; }

	// In world units
	float GetHeight() const { return viewH; }

	// In pixels
	int GetPixWidth() const { return pixW; }

	// In pixels
	int GetPixHeight() const { return pixH; }

	// The minimum zoom-scale value that allows the entire camera world to fit into 
	// the pixel viewport without going past the 2d world limits.
	float GetLimitMinZ() const { return limZ; }

private:
	void Limit_() {
		if (limZ > zoom)
			zoom = limZ;
		if (GetRight() - GetLeft() > limR - limL)
			cx = (limR + limL) * .5f;
		else {
			if (GetLeft() < limL)
				cx = limL + (viewW * .5f) / zoom;
			if (GetRight() > limR)
				cx = limR - (viewW * .5f) / zoom;
		}
		if (GetBottom() - GetTop() > limB - limT)
			cy = (limB + limT) * .5f;
		else {
			if (GetTop() < limT)
				cy = limT + (viewH * .5f) / zoom;
			if (GetBottom() > limB)
				cy = limB - (viewH * .5f) / zoom;
		}
	}

private:
	float zoom;
	float cx, cy;  // world units
	float viewW, viewH;  // world units
	int pixW, pixH;  // pixel units
	float limL, limT, limR, limB, limZ;  // world units
};

}  // namespace glory
