// Color.h - Color objects for easy manipulation of color logic
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v0.99 - 5/19/13 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - Functional and has been used for years. 

//-----------------------------------------------------------------------------------------------------------
// Todo
// - Needs a good refactoring - remove certain non-standard
//   conforming details such as unnamed struct in unnamed union. No real need for Color32::Color. Etc.

#pragma once

#include "Math.h"

namespace glory {

//-----------------------------------------------------------------------------------------------------------
// CLASS PRE-DECLARATIONS (THEY'RE CIRCULARLY DEPENDANT)
//-----------------------------------------------------------------------------------------------------------
class Color32;
class ColorF;

//-----------------------------------------------------------------------------------------------------------
// CLASS DECLARATIONS
//-----------------------------------------------------------------------------------------------------------
class Color32 {
public:
#ifdef COMPILER_MICROSOFT
	#pragma warning(push)
	#pragma warning(disable: 4201)
#endif
	union {
		struct {
			unsigned char B, G, R, A;  // These will define as values of 0xAARRGGBB
		};
		unsigned int Color;
	};
#ifdef COMPILER_MICROSOFT
	#pragma warning(pop)
#endif

#ifdef REBORN_SYS_WINDOWS
	enum WinSystemColor {
		SYSCOLOR_ACTIVEBORDER = COLOR_ACTIVEBORDER,
		SYSCOLOR_ACTIVECAPTION = COLOR_ACTIVECAPTION,
		SYSCOLOR_APPWORKSPACE = COLOR_APPWORKSPACE,
		SYSCOLOR_BACKGROUND = COLOR_BACKGROUND,
		SYSCOLOR_BTNFACE = COLOR_BTNFACE,
		SYSCOLOR_BTNSHADOW = COLOR_BTNSHADOW,
		SYSCOLOR_BTNTEXT = COLOR_BTNTEXT,
		SYSCOLOR_BTNHIGHLIGHT = COLOR_BTNHIGHLIGHT,
		SYSCOLOR_CAPTIONTEXT = COLOR_CAPTIONTEXT,
		SYSCOLOR_GRAYTEXT = COLOR_GRAYTEXT,
		SYSCOLOR_HIGHLIGHT = COLOR_HIGHLIGHT,
		SYSCOLOR_HIGHLIGHTTEXT = COLOR_HIGHLIGHTTEXT,
		SYSCOLOR_INACTIVEBORDER = COLOR_INACTIVEBORDER,
		SYSCOLOR_INACTIVECAPTION = COLOR_INACTIVECAPTION,
		SYSCOLOR_MENU = COLOR_MENU,
		SYSCOLOR_MENUTEXT = COLOR_MENUTEXT,
		SYSCOLOR_SCROLLBAR = COLOR_SCROLLBAR,
		SYSCOLOR_WINDOW = COLOR_WINDOW,
		SYSCOLOR_WINDOWFRAME = COLOR_WINDOWFRAME,
		SYSCOLOR_WINDOWTEXT = COLOR_WINDOWTEXT
	};
#endif

	// Default copy constructor and assignment operators are ok.
	inline Color32();
	inline explicit Color32(unsigned int Value);
	inline Color32(unsigned char Red, unsigned char Green, unsigned char Blue);
	inline Color32(unsigned char Red, unsigned char Green, unsigned char Blue, unsigned char Alpha);
	inline explicit Color32(const ColorF &Value);
#ifdef REBORN_SYS_WINDOWS
	inline explicit Color32(WinSystemColor Color);
	inline explicit Color32(WinSystemColor Color, unsigned char Alpha);
#endif
	inline ~Color32();

	// Casts to a 32 bit color value 0xAARRGGBB
	inline operator unsigned int() const;

	inline bool operator== (const Color32 &Value) const;

	inline Color32 &operator= (unsigned int Value);
	inline Color32 &operator= (const Color32 &Value);

	inline Color32 operator+ (const Color32 &Value) const;
	inline Color32 &operator+= (unsigned int Value);
	inline Color32 operator- (const Color32 &Value) const;
	inline Color32 &operator-= (unsigned int Value);

	inline Color32 operator* (unsigned int Value) const;
	inline Color32 operator* (float Value) const;
	inline Color32 &operator*= (unsigned int Value);
	inline Color32 &operator*= (float Value);

	inline Color32 operator/ (unsigned int Value) const;
	inline Color32 operator/ (float Value) const;
	inline Color32 &operator/= (unsigned int Value);
	inline Color32 &operator/= (float Value);

	// Value of 0xXXBBGGRR or in mem RR GG BB XX
	inline unsigned int GetGDIColor() const;
	// Value of 0xAARRGGBB or in mem BB GG RR AA
	inline unsigned int GetARGB() const;
	// Value of 0xAABBGGRR or in mem RR GG BB AA
	inline unsigned int GetABGR() const;
	// Value of 0xRRGGBBAA or in mem AA BB GG RR
	inline unsigned int GetRGBA() const;
	// Value of 0xBBGGRRAA or in men AA RR GG BB
	inline unsigned int GetBGRA() const;
	// Value of 0xBBGGRR or in mem RR GG BB
	inline unsigned int GetBGR() const;
	// Value of 0xRRGGBB or in mem BB GG RR
	inline unsigned int GetRGB() const;
	// Luminance value from 0 - 255.
	inline unsigned char GetLuminance() const;
	// Average of the values.  Ignores alpha.  Non-destructive.
	inline Color32 GetGray() const;
	// Average of the values.  Ignores alpha.  Destructive.
	inline Color32 &ToGray();
	// Returns the color with each band inverted.  Non-destructive.
	inline Color32 GetInverted() const;
	// Returns the color with each band inverted.  Destructive.
	inline Color32 &Invert();
	// Swaps red and blue channels and returns the color.  Destructive.
	inline Color32 &SwapRB();
	// Swaps red and green channels and returns the color.  Destructive.
	inline Color32 &SwapRG();
	// Swaps blue and green channels and returns the color.  Destructive.
	inline Color32 &SwapGB();
	// Gets the ColorF equivalent
	inline ColorF GetColorF() const;
};

//-----------------------------------------------------------------------------------------------------------
class ColorF {
public:
	float R, G, B, A;
	inline ColorF();
	inline ColorF(float Red, float Green, float Blue);
	inline ColorF(float Red, float Green, float Blue, float Alpha);
	inline explicit ColorF(const Color32 &Value);
	inline ~ColorF();

	friend inline const ColorF operator* (float Val, const ColorF &rhs);
	inline const ColorF operator* (float Val) const;
	inline ColorF &operator*= (float Val);
	inline const ColorF operator/ (float Val) const;
	inline ColorF &operator/= (float Val);

	inline const ColorF operator+ (const ColorF &rhs) const;
	inline ColorF &operator+= (const ColorF &rhs);
	inline const ColorF operator- (const ColorF &rhs) const;
	inline ColorF &operator-= (const ColorF &rhs);

	inline bool operator== (const ColorF &rhs);
	inline bool operator!= (const ColorF &rhs);

	// Luminance value from 0-inf.
	inline float GetLuminance() const;
	// Average of the values.  Ignores alpha.  Non-destructive.
	inline ColorF GetGray() const;
	// Average of the values.  Ignores alpha.  Destructive.
	inline ColorF &ToGray();
	// Returns the color with each band inverted.  Non-destructive.
	inline ColorF GetInverted() const;
	// Returns the color with each band inverted.  Destructive.
	inline ColorF &Invert();
	// Gets the Color32 equivalent.
	inline Color32 GetColor32() const;
};

//-----------------------------------------------------------------------------------------------------------
// CLASS DEFINITIONS
//-----------------------------------------------------------------------------------------------------------
inline Color32::Color32() : Color(0) { }
inline Color32::Color32(unsigned int Value) : Color(Value) { }
inline Color32::Color32(unsigned char Red, unsigned char Green, unsigned char Blue) : B(Blue), G(Green), R(Red), A(255) { }
inline Color32::Color32(unsigned char Red, unsigned char Green, unsigned char Blue, unsigned char Alpha) : B(Blue), G(Green), R(Red), A(Alpha) { }
inline Color32::Color32(const ColorF &Value) { 
	R = static_cast<unsigned char>(GetClamped(Value.R * 255.0f, 0.0f, 255.0f));
	G = static_cast<unsigned char>(GetClamped(Value.G * 255.0f, 0.0f, 255.0f));
	B = static_cast<unsigned char>(GetClamped(Value.B * 255.0f, 0.0f, 255.0f));
	A = static_cast<unsigned char>(GetClamped(Value.A * 255.0f, 0.0f, 255.0f));
}
#ifdef REBORN_SYS_WINDOWS
	inline Color32::Color32(WinSystemColor SysColor) : Color(GetSysColor(SysColor)) { Color = GetBGR(); }
	inline Color32::Color32(WinSystemColor SysColor, unsigned char Alpha) : Color(GetSysColor(SysColor)) { Color = GetBGR(); A = Alpha; }
#endif
inline Color32::~Color32() { }

inline Color32::operator unsigned int() const { 
	return Color;   //return ((A << 24) | (R << 16) | (G << 8) | B); }
}

inline bool Color32::operator== (const Color32 &Value) const {
	return Color == Value.Color;
}

inline Color32 &Color32::operator= (const Color32 &Value) {
	Color = Value.Color;
	return *this;
}

inline Color32 &Color32::operator= (const unsigned int Value) {
	A = Value >> 24;
	R = (Value >> 16) & 0xff;
	G = (Value >> 8) & 0xff;
	B = Value & 0xff;
	return *this;
}

inline Color32 Color32::operator+ (const Color32 &Value) const {
	Color32 RetCol(Value);
	RetCol.R = (RetCol.R + R > 255 ? 255 : RetCol.R + R);
	RetCol.G = (RetCol.G + G > 255 ? 255 : RetCol.G + G);
	RetCol.B = (RetCol.B + B > 255 ? 255 : RetCol.B + B);
	RetCol.A = (RetCol.A + A > 255 ? 255 : RetCol.A + A);
	return RetCol;
}

inline Color32 &Color32::operator+= (unsigned int Value) {
	Color32 RetCol(Value);
	R = (RetCol.R + R > 255 ? 255 : RetCol.R + R);
	G = (RetCol.G + G > 255 ? 255 : RetCol.G + G);
	B = (RetCol.B + B > 255 ? 255 : RetCol.B + B);
	A = (RetCol.A + A > 255 ? 255 : RetCol.A + A);
	return *this;
}

inline Color32 Color32::operator- (const Color32 &Value) const {
	Color32 RetCol(Value);
	RetCol.R = (R - RetCol.R < 0 ? 0 : R - RetCol.R);
	RetCol.G = (G - RetCol.G < 0 ? 0 : G - RetCol.G);
	RetCol.B = (B - RetCol.B < 0 ? 0 : B - RetCol.B);
	RetCol.A = (A - RetCol.A < 0 ? 0 : A - RetCol.A);
	return RetCol;
}

inline Color32 &Color32::operator-= (unsigned int Value) {
	Color32 RetCol(Value);
	R = (R - RetCol.R < 0 ? 0 : R - RetCol.R);
	G = (G - RetCol.G < 0 ? 0 : G - RetCol.G);
	B = (B - RetCol.B < 0 ? 0 : B - RetCol.B);
	A = (A - RetCol.A < 0 ? 0 : A - RetCol.A);
	return *this;
}

inline Color32 Color32::operator* (unsigned int Value) const {
	Color32 RetCol(
		static_cast<unsigned char>(GetClamped<unsigned int>(R * Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(G * Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(B * Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(A * Value, 0, 255)));
	return RetCol;
}
inline Color32 Color32::operator* (float Value) const {
	Color32 RetCol(
		static_cast<unsigned char>(GetClamped(R * Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(G * Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(B * Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(A * Value, 0.0f, 255.0f)));
	return RetCol;
}

inline Color32 &Color32::operator*= (unsigned int Value) {
	R = static_cast<unsigned char>(GetClamped<unsigned int>(R * Value, 0, 255));
	G = static_cast<unsigned char>(GetClamped<unsigned int>(G * Value, 0, 255));
	B = static_cast<unsigned char>(GetClamped<unsigned int>(B * Value, 0, 255));
	A = static_cast<unsigned char>(GetClamped<unsigned int>(A * Value, 0, 255));
	return *this;
}
inline Color32 &Color32::operator*= (float Value) {
	R = static_cast<unsigned char>(GetClamped(R * Value, 0.0f, 255.0f));
	G = static_cast<unsigned char>(GetClamped(G * Value, 0.0f, 255.0f));
	B = static_cast<unsigned char>(GetClamped(B * Value, 0.0f, 255.0f));
	A = static_cast<unsigned char>(GetClamped(A * Value, 0.0f, 255.0f));
	return *this;
}

inline Color32 Color32::operator/ (unsigned int Value) const {
	Color32 RetCol(
		static_cast<unsigned char>(GetClamped<unsigned int>(R / Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(G / Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(B / Value, 0, 255)),
		static_cast<unsigned char>(GetClamped<unsigned int>(A / Value, 0, 255)));
	return RetCol;
}
inline Color32 Color32::operator/ (float Value) const {
	Color32 RetCol(
		static_cast<unsigned char>(GetClamped(R / Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(G / Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(B / Value, 0.0f, 255.0f)),
		static_cast<unsigned char>(GetClamped(A / Value, 0.0f, 255.0f)));
	return RetCol;
}

inline Color32 &Color32::operator/= (unsigned int Value) {
	R = static_cast<unsigned char>(GetClamped<unsigned int>(R / Value, 0, 255));
	G = static_cast<unsigned char>(GetClamped<unsigned int>(G / Value, 0, 255));
	B = static_cast<unsigned char>(GetClamped<unsigned int>(B / Value, 0, 255));
	A = static_cast<unsigned char>(GetClamped<unsigned int>(A / Value, 0, 255));
	return *this;
}
inline Color32 &Color32::operator/= (float Value) {
	R = static_cast<unsigned char>(GetClamped(R / Value, 0.0f, 255.0f));
	G = static_cast<unsigned char>(GetClamped(G / Value, 0.0f, 255.0f));
	B = static_cast<unsigned char>(GetClamped(B / Value, 0.0f, 255.0f));
	A = static_cast<unsigned char>(GetClamped(A / Value, 0.0f, 255.0f));
	return *this;
}

inline unsigned int Color32::GetABGR() const {
	return ((B << 16) | (R) | (Color & 0xff00ff00));
}

inline unsigned int Color32::GetARGB() const {
	return Color;
}

inline unsigned int Color32::GetBGRA() const {
	return (A | (R << 8) | (G << 16) | (B << 24));
}

inline unsigned int Color32::GetRGBA() const {
	return (A | (B << 8) | (G << 16) | (R << 24));
}

inline unsigned int Color32::GetGDIColor() const {
	// Alpha component actually makes GDI render all 0s!!
	return GetBGR();
}

inline unsigned int Color32::GetBGR() const {
	return ((B << 16) | (R) | (Color & 0x0000ff00));
}

inline unsigned int Color32::GetRGB() const {
	return Color & 0x00ffffff;
}

inline unsigned char Color32::GetLuminance() const {
	return static_cast<unsigned char>(R * .299f + G * .587f + B * 0.114f);
}

inline Color32 Color32::GetGray() const { 
	unsigned char Avg = ((unsigned int)R + (unsigned int)G + (unsigned int)B) / 3; 
	return Color32(Avg, Avg, Avg);
}

inline Color32 &Color32::ToGray() { 
	return (*this = GetGray()); 
}

inline Color32 Color32::GetInverted() const {
	return Color32(255 - R, 255 - G, 255 - B, 255 - A);
}

inline Color32 &Color32::Invert() {
	R = 255 - R; G = 255 - G; B = 255 - B; A = 255 - A;
	return *this;
}

inline Color32 &Color32::SwapRB() { 
	unsigned char T = R; R = B; B = T; return *this; 
}

inline Color32 &Color32::SwapRG() { 
	unsigned char T = R; R = G; G = T; return *this; 
}

inline Color32 &Color32::SwapGB() { 
	unsigned char T = G; G = B; B = T; return *this; 
}

inline ColorF Color32::GetColorF() const {
	return ColorF(*this);
}

//-----------------------------------------------------------------------------------------------------------
inline ColorF::ColorF() : R(0), G(0), B(0), A(0) { }
inline ColorF::ColorF(float Red, float Green, float Blue) : R(Red), G(Green), B(Blue), A(1.0f) { }
inline ColorF::ColorF(float Red, float Green, float Blue, float Alpha) : R(Red), G(Green), B(Blue), A(Alpha) { }
inline ColorF::ColorF(const Color32 &Value) {
	R = static_cast<float>(Value.R) / 255.0f;
	G = static_cast<float>(Value.G) / 255.0f;
	B = static_cast<float>(Value.B) / 255.0f;
	A = static_cast<float>(Value.A) / 255.0f;
}
inline ColorF::~ColorF() { }

inline const ColorF operator *(float Val, const ColorF &rhs) {
	return ColorF(rhs.R * Val, rhs.G * Val, rhs.B * Val, rhs.A * Val); 
}

inline const ColorF ColorF::operator* (float Val) const { 
	return ColorF(R * Val, G * Val, B * Val, A * Val); 
}

inline ColorF &ColorF::operator*= (float Val) {
	R *= Val;
	G *= Val;
	B *= Val;
	A *= Val;
	return *this;
}

inline const ColorF ColorF::operator/ (float Val) const { 
	return ColorF(R / Val, G / Val, B / Val, A / Val); 
}

inline ColorF &ColorF::operator/= (float Val) {
	R /= Val;
	G /= Val;
	B /= Val;
	A /= Val;
	return *this;
}

inline const ColorF ColorF::operator+ (const ColorF &rhs) const {
	return ColorF(R + rhs.R, G + rhs.G, B + rhs.B, A + rhs.A);
}

inline ColorF &ColorF::operator+= (const ColorF &rhs) {
	R += rhs.R;
	G += rhs.G;
	B += rhs.B;
	A += rhs.A;
	return *this;
}

inline const ColorF ColorF::operator- (const ColorF &rhs) const {
	return ColorF(R - rhs.R, G - rhs.G, B - rhs.B, A - rhs.A);
}

inline ColorF &ColorF::operator-= (const ColorF &rhs) {
	R -= rhs.R;
	G -= rhs.G;
	B -= rhs.B;
	A -= rhs.A;
	return *this;
}

inline bool ColorF::operator== (const ColorF &rhs) {
	return (FuzzyEq(R, rhs.R) && 
		FuzzyEq(G, rhs.G) && 
		FuzzyEq(B, rhs.B) && 
		FuzzyEq(A, rhs.A));
}

inline bool ColorF::operator!= (const ColorF &rhs) {
	return !(*this == rhs);
}

inline float ColorF::GetLuminance() const {
	return R * .299f + G * .587f + B * 0.114f;
}

inline ColorF ColorF::GetGray() const { 
	float Avg = (R + G + B) / 3.0f; 
	return ColorF(Avg, Avg, Avg);
}

inline ColorF &ColorF::ToGray() { 
	return (*this = GetGray()); 
}

inline ColorF ColorF::GetInverted() const {
	return ColorF(1 - R, 1 - G, 1 - B, 1 - A);
}

inline ColorF &ColorF::Invert() {
	R = 1 - R; G = 1 - G; B = 1 - B; A = 1 - A;
	return *this;
}

inline Color32 ColorF::GetColor32() const {
	return Color32(*this);
}

//-----------------------------------------------------------------------------------------------------------
// Utility function specializations
//-----------------------------------------------------------------------------------------------------------
inline Color32 GetClamped(Color32 A, Color32 Min, Color32 Max) { 
	return Color32(GetClamped(A.R, Min.R, Max.R), 
		GetClamped(A.G, Min.G, Max.G), 
		GetClamped(A.B, Min.B, Max.B), 
		GetClamped(A.A, Min.A, Max.A));
}
inline ColorF GetClamped(ColorF A, ColorF Min, ColorF Max) { 
	return ColorF(GetClamped(A.R, Min.R, Max.R), 
		GetClamped(A.G, Min.G, Max.G), 
		GetClamped(A.B, Min.B, Max.B), 
		GetClamped(A.A, Min.A, Max.A));
}
inline void AssignMinMax(Color32 &CurrentMin, Color32 &CurrentMax, const Color32 &Value) {
	AssignMinMax(CurrentMin.R, CurrentMax.R, Value.R);
	AssignMinMax(CurrentMin.G, CurrentMax.G, Value.G);
	AssignMinMax(CurrentMin.B, CurrentMax.B, Value.B);
	AssignMinMax(CurrentMin.A, CurrentMax.A, Value.A);
}
inline void AssignMinMax(ColorF &CurrentMin, ColorF &CurrentMax, const ColorF &Value) {
	AssignMinMax(CurrentMin.R, CurrentMax.R, Value.R);
	AssignMinMax(CurrentMin.G, CurrentMax.G, Value.G);
	AssignMinMax(CurrentMin.B, CurrentMax.B, Value.B);
	AssignMinMax(CurrentMin.A, CurrentMax.A, Value.A);
}

}  // namespace glory
