// Rect.h - Common rectangle math routines.
// This is especially useful when it comes to displaying images or working
// out bounds in a GUI environment.
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.02 - 04/22/13 - Fixed rare case in FitCentered
// - v1.01 - 09/15/12 - Added IsRectIn()
// - v1.00 - 05/31/12 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - Unlike windows' RECT, this is inclusive, not exclusive.
//   Meaning: This rectangle class isn't mathematically accurate, but it does work nicely
//   for 2d integer pixel grids.  Note that this isn't a replacement 2d AABB.
// - Can be very confusing when mixing with sub-pixel accurate coordinate systems. If this
//   is the case, using right and bottom exclusive integer coordinates such as how
//   windows handles RECTs makes much more mathematical sense.

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

namespace glory {

class Rect {
public:
	int l, t, r, b;

	Rect() : l(0), t(0), r(0), b(0) { }
	Rect(int w, int h) : l(0), t(0), r(w - 1), b(h - 1) { }
	Rect(int l, int t, int r, int b) : l(l), t(t), r(r), b(b) { }
	Rect(const Rect &copy) : l(copy.l), t(copy.t), r(copy.r), b(copy.b) { }

	Rect &operator= (const Rect &copy) { l = copy.l; t = copy.t; r = copy.r; b = copy.b; return *this; }

#ifdef _MSC_VER
	Rect(const RECT &winRect) : l(winRect.left), t(winRect.top), r(winRect.right - 1), b(winRect.bottom - 1) { }
	Rect(const POINT &winPtLT, const POINT &winPtRB) : l(winPtLT.x), t(winPtLT.y), r(winPtRB.x), b(winPtRB.y) { }

	Rect &operator= (const RECT &winRect) { l = winRect.left; t = winRect.top; r = winRect.right - 1; b = winRect.bottom - 1; return *this; }

	RECT GetRECT() const {
		RECT rc;
		// Window's RECT is bottom-right exclusive
		rc.left = l;  rc.top = t;  rc.right = r + 1;  rc.bottom = b + 1;
		return rc;
	}
#endif

	int GetWidth() const { return r - l + 1; }
	int GetHeight() const { return b - t + 1; }
	int GetCenterX() const { return (l + r) >> 1; }
	int GetCenterY() const { return (t + b) >> 1; }

	void Offset(int x, int y) { l += x; r += x; t += y; b += y; }
	void Adjust(int la, int ta, int ra, int ba) { l += la; t += ta; b += ba; r += ra; }

	bool IsPointIn(int x, int y) { return x >= l && x <= r && y >= t && y <= b; }
	bool IsRectIn(const Rect &other) { return other.l <= l && other.r >= r && other.b >= b && other.t <= t; }

	// Clips against the bounds
	void Clip(const Rect &bounds) {
		if (r > bounds.r)
			r = bounds.r;
		if (r < bounds.l)
			r = l - 1;
		if (l < bounds.l)
			l = bounds.l;
		if (l > bounds.r)
			l = r + 1;
		if (b > bounds.b)
			b = bounds.b;
		if (b < bounds.t)
			b = t - 1;
		if (t < bounds.t)
			t = bounds.t;
		if (t > bounds.b)
			t = b + 1;
	}

	// Clips based on size only.  Only modifies r and b of self.
	void ClipSize(const Rect &bounds) {
		if (GetWidth() > bounds.GetWidth())
			r = l + bounds.GetWidth() - 1;
		if (GetHeight() > bounds.GetHeight())
			b = t + bounds.GetHeight() - 1;
	}

	// Clips based on size.  Only modifies r and b of self.  Maintain's ratio.
	// Ex. Fitting an image into a window from top left corner without scaling up if smaller.
	void ClipSizeRatio(const Rect &bounds) {
		int w = GetWidth();
		int h = GetHeight();
		int bw = bounds.GetWidth();
		int bh = bounds.GetHeight();
		if (h > bh) {
			w = w * bh / h;  // resize width by the same ratio that the height will resize
			h = bh;  // h * bh / h
		}
		if (w > bw) {
			h = h * bw / w;  // resize height by the same ratio that the width will resize
			w = bw;  // w * bw / w
		}
		r = l + w - 1;
		b = t + h - 1;
	}

	// Ex. Keeping a movable box or a selection box within bounds.
	void ClampAndClip(const Rect &bounds) {
		if (r > bounds.r)
			Offset(bounds.r - r, 0);
		if (l < bounds.l)
			Offset(bounds.l - l, 0);
		if (b > bounds.b)
			Offset(0, bounds.b - b);
		if (t < bounds.t)
			Offset(0, bounds.t - t);
		Clip(bounds);
	}

	// Expand if smaller, detract if larger, and center inside the bounds.
	// Ex. an image preview that should always scale to the preview box.
	void FitCentered(const Rect &bounds) {
		int w = GetWidth();
		int h = GetHeight();
		int bw = bounds.GetWidth();
		int bh = bounds.GetHeight();
		// @@ Better to test which one of these will suffice instead of
		// going through both.  50% of time, both will happen, not just 1,
		// and the first one will be pointless in this case.  Once we find the test,
		// the "h = bh" line will be pointless as well.
		if (h != bh) {
			w = w * bh / h;  // resize width by the same ratio that the height will resize
			h = bh;  // h * bh / h - still need it in case next conditional happens
			t = bounds.t;
			b = bounds.b;
			l = (bounds.r + bounds.l - w) / 2;
			r = l + w;
		}
		else if (w < bw) {
			// Get here in rare cases, but it can happen. h == bh, and w < bw
			l = (bounds.r + bounds.l - w) / 2;
			r = l + w;
		}
		if (w > bw) {
			h = h * bw / w;  // resize height by the same ratio that the width will resize
			l = bounds.l;
			r = bounds.r;
			t = (bounds.b + bounds.t - h) / 2;
			b = t + h;
		}
	}

	// Center the rectangle inside or around another - no scaling this time.
	void Center(const Rect &bounds) {
		Offset(bounds.l + (bounds.GetWidth() - GetWidth()) / 2 - l,
			bounds.t + (bounds.GetHeight() - GetHeight()) / 2 - t);
	}
};

}  // namespace glory
