// TexTri32.h - Generic, but sub-pixel and sub-texel accurate,
// affine mapper for 32 bits 0xaarrggbb or 0xaabbggrr
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.00 - 07/05/13 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - Typically, you'll only include this in one place that uses/wraps it
// - Define TEXTRI32_DECLARE_ONLY to not include the definition
// - Not highly optimized, but makes a very great reference or starting place
//   for more advanced rasterizers
// - Input and output pixel format are expected to match and have alpha in high bits
// - Unlike most of glory, this is fully usable in C

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#pragma once

#ifndef TEXTRI32_H_INCLUDED
#define TEXTRI32_H_INCLUDED

#ifdef TEXTRI32_C_ONLY
namespace glory {
#endif

typedef struct TexVert {
    float x, y, u, v;
} TexVert;

#define f16ceili(a) (((a) + 0xffff) >> 16)
#define f16ceilf16(a) (((a) + 0xffff) & 0xffff0000)
#define f16toi(a) ((a) >> 16)
#define ftof16(a) ((int)((a) * 65536.0f))
// Barely nudge up on the left sides
#define f16mulf16(a, b) (((a + 0xff) >> 8) * ((b + 0xff) >> 8))
// Barely nudge down on the right sides
#define f16mullowf16(a, b) (((a) >> 8) * ((b) >> 8))

// spans are in pixels
#ifdef TEXTRI32_DECLARE_ONLY
void DrawTri(unsigned int *pData, int w, int h, int rgbspan, unsigned int *pTexData, int tspan, TexVert *v);
#else
void DrawTri(unsigned int *pData, int w, int h, int rgbspan, unsigned int *pTexData, int tspan, TexVert *v) {
    TexVert *v1 = v, *v2 = v + 1, *v3 = v + 2;
    TexVert *vt;
    int yi, ys, ye;
    int du, dv;
    int ul = 0, vl = 0, ui, vi;
    int dydxl = 0, dydxr = 0, dydu = 0, dydv = 0;
    int xi, xs, xe, xl = 0, xr = 0;
	unsigned int *pOut;
	unsigned int pix;
	unsigned char a, ia, ca;
	int pre;

	int fpy1, fpy2, fpx1, fpx2;
	int sech;
    float d;

    // Sort by y
    if (v2->y < v1->y) { vt = v1; v1 = v2; v2 = vt; }
    if (v3->y < v1->y) { vt = v1; v1 = v3; v3 = vt; }
    if (v3->y < v2->y) { vt = v2; v2 = v3; v3 = vt; }
    
	d = (v1->x - v3->x) * (v2->y - v3->y) - (v2->x - v3->x) * (v1->y - v3->y);
	if (d == 0)
		return;
	d = 1.0f / d * 65536.0f;
	du = (int)(((v1->u - v3->u) * (v2->y - v3->y) - (v2->u - v3->u) * (v1->y - v3->y)) * d);
	dv = (int)(((v1->v - v3->v) * (v2->y - v3->y) - (v2->v - v3->v) * (v1->y - v3->y)) * d);
	fpx1 = ftof16(v1->x);
	fpx2 = ftof16(v2->x);
	fpy1 = ftof16(v1->y);
	fpy2 = ftof16(v2->y);

	ys = f16ceili(fpy1);
    ye = f16ceili(fpy2);
	sech = ye - ys;
    if (sech > 0) {
    	if (d < 0) {
    		dydxl = ftof16((v2->x - v1->x) / (v2->y - v1->y));
    		dydu = ftof16((v2->u - v1->u) / (v2->y - v1->y));
    		dydv = ftof16((v2->v - v1->v) / (v2->y - v1->y));
    		dydxr = ftof16((v3->x - v1->x) / (v3->y - v1->y));
    	}
    	else {
    		dydxl = ftof16((v3->x - v1->x) / (v3->y - v1->y));
    		dydu = ftof16((v3->u - v1->u) / (v3->y - v1->y));
    		dydv = ftof16((v3->v - v1->v) / (v3->y - v1->y));
    		dydxr = ftof16((v2->x - v1->x) / (v2->y - v1->y));
    	}

		pre = f16ceilf16(fpy1) - fpy1;
    	xl = fpx1 + f16mulf16(pre, dydxl);
    	ul = ftof16(v1->u) + f16mulf16(pre, dydu);
    	vl = ftof16(v1->v) + f16mulf16(pre, dydv);
    	xr = fpx1 + f16mullowf16(pre, dydxr);
        for (yi = ys; yi < ye; yi++) {
   			xs = f16ceili(xl);
   			xe = f16ceili(xr);
            if (yi >= 0 && yi < h && xe >= 0 && xs < w) {
    			pre = f16ceilf16(xl) - xl;
				ui = ul + f16mulf16(pre, du);
    			vi = vl + f16mulf16(pre, dv);
    			if (xs < 0) {
                    ui += du * -xs;
                    vi += dv * -xs;
                    xs = 0;
                }
                if (xe >= w)
                    xe = w - 1;
    	        pOut = pData + yi * rgbspan + xs;
				
                for (xi = xs; xi < xe; xi++) {
    				pix = pTexData[f16toi(vi) * tspan + f16toi(ui)];
					a = pix >> 24;
					if (a == 0xff)
						*pOut = pix;
					else if (a) {
						ca = *pOut >> 24;
						ia = ((0xff - a) * ca) >> 8;
						ca = (ca + a) > 0xff ? 0xff : (ca + a);
						pix = (((((*pOut & 0x00ff00ff) * ia) & 0xff00ff00) + (((pix & 0x00ff00ff) * a) & 0xff00ff00)) >> 8)
							| (((((*pOut & 0xff00) * ia) & 0xff0000) + (((pix & 0xff00) * a) & 0xff0000)) >> 8)
							| (ca << 24);
						*pOut = pix;
					}
    				pOut++;
    				ui += du;
    				vi += dv;
    			}
    		}
            xl += dydxl;
            ul += dydu;
            vl += dydv;
    		xr += dydxr;
        }
    }
    else {
		pre = f16ceilf16(fpy1) - fpy1;
        if (d < 0) {
    		dydxr = ftof16((v3->x - v1->x) / (v3->y - v1->y));
			xr = fpx1 + f16mullowf16(pre, dydxr);
		}
        else {
    		dydxl = ftof16((v3->x - v1->x) / (v3->y - v1->y));
    		dydu = ftof16((v3->u - v1->u) / (v3->y - v1->y));
    		dydv = ftof16((v3->v - v1->v) / (v3->y - v1->y));
			xl = fpx1 + f16mulf16(pre, dydxl);
    		ul = ftof16(v1->u) + f16mulf16(pre, dydu);
    		vl = ftof16(v1->v) + f16mulf16(pre, dydv);
		}
    }

	ys = f16ceili(fpy2);
    ye = f16ceili(ftof16(v3->y));
	sech = ye - ys;
    if (sech > 0) {

		pre = f16ceilf16(fpy2) - fpy2;

		if (d < 0) {
    		dydxl = ftof16((v3->x - v2->x) / (v3->y - v2->y));
    		dydu = ftof16((v3->u - v2->u) / (v3->y - v2->y));
    		dydv = ftof16((v3->v - v2->v) / (v3->y - v2->y));
			xl = fpx2 + f16mulf16(pre, dydxl);
    		ul = ftof16(v2->u) + f16mulf16(pre, dydu);
    		vl = ftof16(v2->v) + f16mulf16(pre, dydv);
    	}
    	else {
    		dydxr = ftof16((v3->x - v2->x) / (v3->y - v2->y));
			xr = fpx2 + f16mullowf16(pre, dydxr);
    	}

        for (yi = ys; yi < ye; yi++) {
   			xs = f16ceili(xl);
   			xe = f16ceili(xr);
            if (yi >= 0 && yi < h && xe >= 0 && xs < w) {
    			pre = f16ceilf16(xl) - xl;
				ui = ul + f16mulf16(pre, du);
    			vi = vl + f16mulf16(pre, dv);
    			if (xs < 0) {
                    ui += du * -xs;
                    vi += dv * -xs;
                    xs = 0;
                }
                if (xe >= w)
                    xe = w - 1;
    	        pOut = pData + yi * rgbspan + xs;
				
                for (xi = xs; xi < xe; xi++) {
    				pix = pTexData[f16toi(vi) * tspan + f16toi(ui)];
					a = pix >> 24;
					if (a == 0xff)
						*pOut = pix;
					else if (a) {
						ca = *pOut >> 24;
						ia = ((0xff - a) * ca) >> 8;
						ca = (ca + a) > 0xff ? 0xff : (ca + a);
						pix = (((((*pOut & 0x00ff00ff) * ia) & 0xff00ff00) + (((pix & 0x00ff00ff) * a) & 0xff00ff00)) >> 8)
							| (((((*pOut & 0xff00) * ia) & 0xff0000) + (((pix & 0xff00) * a) & 0xff0000)) >> 8)
							| (ca << 24);
						*pOut = pix;
					}
    				pOut++;
    				ui += du;
    				vi += dv;
    			}
    		}
            xl += dydxl;
            ul += dydu;
            vl += dydv;
    		xr += dydxr;
        }
    }
}
#endif

#undef f16ceili
#undef f16ceilf16
#undef f16toi
#undef ftof16
#undef f16mulf16
#undef f16mullowf16

#ifdef TEXTRI32_C_ONLY
}  // namespace glory
#endif

#endif //TEXTRI32_H_INCLUDED
