// Vector.h - Vector class. These are a dime-a-dozen.
//
// This is free and unencumbered software released into the public domain.
// 
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>

//-----------------------------------------------------------------------------------------------------------
// History
// - v1.03 - 10/06/12 - Fixed bugs in Vector2::Project, Rotate, and Unproject
// - v1.02 - 09/15/12 - Added many functions to both Vector2 and Vector3
// - v1.01 - 06/07/12 - Corrected namespace to "glory"
// - v1.00 - 05/31/12 - Initial release by achild

//-----------------------------------------------------------------------------------------------------------
// Notes
// - 

//-----------------------------------------------------------------------------------------------------------
// Todo
// - 

#include "Math.h"

#pragma once

namespace glory {

#pragma pack(push, 4)
//-----------------------------------------------------------------------------------------------------------
class Vector2 {
public:
	float X, Y;
	Vector2() : X(0), Y(0) { }
	Vector2(const Vector2 &Copy) : X(Copy.X), Y(Copy.Y) { }
	Vector2(float X, float Y) : X(X), Y(Y) { }
	inline Vector2 &operator= (const Vector2 &rhs) { X = rhs.X; Y = rhs.Y; return *this; }

	inline bool operator== (const Vector2 &rhs) const {
		return FuzzyEq(X, rhs.X) && FuzzyEq(Y, rhs.Y);
	}

	inline bool operator!= (const Vector2 &rhs) const {
		return !(*this == rhs);
	}

	inline Vector2 operator+ (const Vector2 &rhs) const { return Vector2(X + rhs.X, Y + rhs.Y); }

	inline Vector2 &operator+= (const Vector2 &rhs) {
		X += rhs.X;
		Y += rhs.Y;
		return *this;
	}

	inline Vector2 operator- (const Vector2 &rhs) const { return Vector2(X - rhs.X, Y - rhs.Y); }

	inline Vector2 &operator-= (const Vector2 &rhs) {
		X -= rhs.X;
		Y -= rhs.Y;
		return *this;
	}
	inline Vector2 operator* (float rhs) const { return Vector2(X * rhs, Y * rhs); }
	inline Vector2 &operator*= (float rhs) { X *= rhs; Y *= rhs; return *this; }
	friend inline Vector2 operator* (float lhs, const Vector2 &rhs) { 
		return Vector2(rhs.X * lhs, rhs.Y * lhs); 
	}

	// Should this check for a value of 0?
	inline Vector2 operator/ (float rhs) const { return Vector2(X / rhs, Y / rhs); }
	inline Vector2 &operator/= (float rhs) { X /= rhs; Y /= rhs; return *this; }

	void Normalize() {
		float Mag2 = X * X + Y * Y;
		if (Mag2 > .00001) {
			float Mag = sqrt(Mag2);
			X /= Mag;
			Y /= Mag;
		}
	}

	Vector2 GetUnit() const {
		float Mag2 = X * X + Y * Y;
		if (Mag2 > .00001) {
			float Mag = sqrt(Mag2);
			return Vector2(X / Mag, Y / Mag);
		}
		return Vector2(X, Y);
	}

	void Rotate(float Rad) {
		float cosRad = cos(Rad);
		float sinRad = sin(Rad);
		float oldX = X;
		X = X * cosRad - Y * sinRad;
		Y = Y * cosRad + oldX * sinRad;	
	}

	Vector2 GetRotated(float Rad) {
		float cosRad = cos(Rad);
		float sinRad = sin(Rad);
		return Vector2(X * cosRad - Y * sinRad, Y * cosRad + X * sinRad);
	}

	void Reflect(const Vector2 &NormVec) {
		float Dot = X * NormVec.X + Y * NormVec.Y;
		X -= 2 * NormVec.X * Dot;
		Y -= 2 * NormVec.Y * Dot;
	}

	Vector2 GetReflected(const Vector2 &NormVec) const {
		float Dot = X * NormVec.X + Y * NormVec.Y;
		return Vector2(X - 2 * NormVec.X * Dot, Y - 2 * NormVec.Y * Dot);
	}

	// Go from being axis aligned into a certain orientation. Or - takes vector from orientation's local space.
	void Project(const Vector2 &Orient) {
		float oldX = X;
		X = X * Orient.X - Y * Orient.Y;
		Y = oldX * Orient.Y + Y * Orient.X;
	}

	Vector2 GetProjected(const Vector2 &Orient) {
		return Vector2(X * Orient.X - Y * Orient.Y, X * Orient.Y + Y * Orient.X);
	}

	// Go from a certain orientation into being axis aligned. Or - brings vector into orientation's local space.
	void Unproject(const Vector2 &Orient) {
		float oldX = X;
		X = X * Orient.X + Y * Orient.Y;
		Y = Y * Orient.X - oldX * Orient.Y;
	}
	
	Vector2 GetUnprojected(const Vector2 &Orient) {
		return Vector2(X * Orient.X + Y * Orient.Y, Y * Orient.X - X * Orient.Y);
	}

	float GetLength() const {
		return sqrt(X * X + Y * Y);
	}

	float GetLengthSqr() const {
		return X * X + Y * Y;
	}

	// Angle between the vector going from (0, 0) to (X, Y), and vector going from (0, 0) to (1, 0)
	float GetAngleRad() const {
		return atan2(Y, X);
	}

	float GetAngleRadFrom(const Vector2 &VecFrom) const {
		// Unproject this vector from VecFrom's local space, then get the angle the normal way
		return atan2(Y * VecFrom.X - X * VecFrom.Y, X * VecFrom.X + Y * VecFrom.Y);
	}

	float GetDot(const Vector2 &Other) {
		return X * Other.X + Y * Other.Y;
	}

	bool IsClockwiseFrom(const Vector2 &VecFrom) const {
		float Z = X * VecFrom.Y - Y * VecFrom.X;
		return Z > 0;
	}

	// Premade vectors
	static inline Vector2 UnitX() { return Vector2(1, 0); }
	static inline Vector2 UnitY() { return Vector2(0, 1); }
};

//-----------------------------------------------------------------------------------------------------------
class Vector3 {
public:
	float X, Y, Z;
	Vector3() : X(0), Y(0), Z(0) { }
	Vector3(const Vector3 &Copy) : X(Copy.X), Y(Copy.Y), Z(Copy.Z) { }
	Vector3(float X, float Y, float Z) : X(X), Y(Y), Z(Z) { }
	inline Vector3 &operator= (const Vector3 &rhs) { X = rhs.X; Y = rhs.Y; Z = rhs.Z; return *this; }

	inline bool operator== (const Vector3 &rhs) const {
		return FuzzyEq(X, rhs.X) && FuzzyEq(Y, rhs.Y) && FuzzyEq(Z, rhs.Z);
	}

	inline bool operator!= (const Vector3 &rhs) const {
		return !(*this == rhs);
	}

	inline Vector3 operator+ (const Vector3 &rhs) const { return Vector3(X + rhs.X, Y + rhs.Y, Z + rhs.Z); }

	inline Vector3 &operator+= (const Vector3 &rhs) {
		X += rhs.X;
		Y += rhs.Y;
		Z += rhs.Z;
		return *this;
	}

	inline Vector3 operator- (const Vector3 &rhs) const { return Vector3(X - rhs.X, Y - rhs.Y, Z - rhs.Z); }

	inline Vector3 &operator-= (const Vector3 &rhs) {
		X -= rhs.X;
		Y -= rhs.Y;
		Z -= rhs.Z;
		return *this;
	}

	inline Vector3 operator* (float rhs) const { return Vector3(X * rhs, Y * rhs, Z * rhs); }
	inline Vector3 &operator*= (float rhs) { X *= rhs; Y *= rhs; Z *= rhs; return *this; }
	friend inline const Vector3 operator* (float lhs, const Vector3 &rhs) { 
		return Vector3(rhs.X * lhs, rhs.Y * lhs, rhs.Z * lhs); 
	}

	// Should this check for a value of 0?
	inline Vector3 operator/ (float rhs) const { return Vector3(X / rhs, Y / rhs, Z / rhs); }
	inline Vector3 &operator/= (float rhs) { X /= rhs; Y /= rhs; Z /= rhs; return *this; }


	void Normalize() {
		float Mag2 = X * X + Y * Y + Z * Z;
		if (Mag2 > .00001) {
			float Mag = sqrt(Mag2);
			X /= Mag;
			Y /= Mag;
			Z /= Mag;
		}
	}

	Vector3 GetUnit() const {
		float Mag2 = X * X + Y * Y + Z * Z;
		if (Mag2 > .00001) {
			float Mag = sqrt(Mag2);
			return Vector3(X / Mag, Y / Mag, Z / Mag);
		}
		return Vector3(X, Y, Z);
	}

	void Reflect(const Vector3 &NormVec) {
		float Dot = X * NormVec.X + Y * NormVec.Y + Z * NormVec.Z;
		X -= 2 * NormVec.X * Dot;
		Y -= 2 * NormVec.Y * Dot;
		Z -= 2 * NormVec.Z * Dot;
	}

	Vector3 GetReflected(const Vector3 &NormVec) const {
		float Dot = X * NormVec.X + Y * NormVec.Y + Z * NormVec.Z;
		return Vector3(X - 2 * NormVec.X * Dot, Y - 2 * NormVec.Y * Dot, Z - 2 * NormVec.Z * Dot);
	}

	float GetLength() const {
		return sqrt(X * X + Y * Y + Z * Z);
	}

	float GetLengthSqr() const {
		return X * X + Y * Y + Z * Z;
	}

	// this * that
	// Get the cross product of 2 sides of a triangle,
	// normalize the result, and you have the normal vector for that triangle.
	inline Vector3 GetCross(const Vector3 &rhs) const {
		return Vector3(Y * rhs.Z - rhs.Y * Z, Z * rhs.X - rhs.Z * X, X * rhs.Y - rhs.X * Y);
	}

	// Normalize both before getting the dot product, which is the cosine of the angle between v1 and v2
	// acos of this gives the angle 0 - Pi
	inline float GetDot(const Vector3 &rhs) const {
		return X * rhs.X + Y * rhs.Y + Z * rhs.Z;
	}

	// Make sure they are both normalized...
	// To get a quaternion, xyz = the cross product, and w = 1 + the dot product, if they are both normalized...
	inline void GetAxisAngleTo(const Vector3 &ToThis, Vector3 &OutAxis, float &OutRadian) const {
		OutRadian = acos(GetDot(ToThis));
		OutAxis = GetCross(ToThis).GetUnit();
	}

	// Premade vectors
	static inline Vector3 UnitX() { return Vector3(1, 0, 0); }
	static inline Vector3 UnitY() { return Vector3(0, 1, 0); }
	static inline Vector3 UnitZ() { return Vector3(0, 0, 1); }
};
#pragma pack(pop)

}  // namespace glory
